"""
europilot.screen
~~~~~~~~~~~~~~~~

This module contains wrappers which help to read screen data from
local/remote machine.

"""

import os
import ast
import time
import traceback
from sys import platform
from itertools import count
from subprocess import Popen, PIPE

import numpy as np
from mss import mss

from europilot.exceptions import ScreenException


class Box(object):
    # Global counter which is used to name box
    _counter = count()

    def __init__(self, x1, y1, x2, y2, monitor=None):
        """Constructor

        (x1, y1) -> upper left point of the screen box
        (x2, y2) -> lower right point of the screen box

              |
              |
        ---(x1, y1)-----------------(x2, y1)
              |                        |
              |                        |
              |                        |
              |                        |
        ---(x1, y2)-----------------(x2, y2)

        :param monitor: `Monitor` object which adds proper offset
        according to target minitor.

        """
        self._name = 'Box %s' % str(next(self._counter))
        self._x1 = x1
        self._y1 = y1
        self._x2 = x2
        self._y2 = y2
        self._monitor = monitor

        if self._monitor is not None:
            # TODO: This is needed to support multiple monitors.
            pass

    @staticmethod
    def from_monitor(monitor):
        return Box(
            monitor.offset_x, monitor.offset_y,
            monitor.width, monitor.height
        )

    @staticmethod
    def from_tuple(tuple_):
        return Box(
            tuple_[0],
            tuple_[1],
            tuple_[2],
            tuple_[3]
        )

    @property
    def name(self):
        return self._name

    @property
    def width(self):
        return self._x2 - self._x1

    @property
    def height(self):
        return self._y2 - self._y1

    @property
    def channel(self):
        # Channel for RGB
        return 3

    @property
    def numpy_shape(self):
        return (self.height, self.width, self.channel)

    def to_tuple(self):
        return (self._x1, self._y1, self._x2, self._y2)


class Monitor(object):
    """This class holds monitor information.

    """
    def __init__(self, width, height, offset_x, offset_y, primary=False):
        """Constructor

        :param width: Monitor width
        :param height: Monitor height
        :param offset_x, offset_y:
        If there are multiple monitors, other monitors expect for primary
        one have a offsets according to os display config.
        :param primary: Whether it is a primary monitor or not.

        """
        self._width = width
        self._height = height
        self._offset_x = offset_x
        self._offset_y = offset_y
        self._primary = primary

    @property
    def primary(self):
        return self._primary

    @property
    def width(self):
        return self._width

    @property
    def height(self):
        return self._height

    @property
    def offset_x(self):
        return self._offset_x

    @property
    def offset_y(self):
        return self._offset_y


class ScreenUtils(object):
    @staticmethod
    def select_screen_area():
        """Use opencv to select game window from entire screen and return
        `box` object corresponding to game window.
        This method uses `selectROI` method in opencv tracking api.
        See http://docs.opencv.org/master/d7/dfc/group__highgui.html
        NOTE that opencv 3.0 (or above) is required.

        """
        try:
            import cv2
            cv2.selectROI
        except ImportError:
            raise ScreenException('opencv is not Found')
        except AttributeError:
            raise ScreenException('`selectROI` not found.' +
                ' Try reinstalling opencv with `--with-contrib` option')

        # 1. Capture entire screen in primary monitor.
        monitors = ScreenUtils.get_local_monitors()
        # Use primary monitor to create box
        box = Box.from_monitor(monitors[0])
        local_grab = LocalScreenGrab(box)
        entire_screen = local_grab.grab()
        entire_screen = entire_screen.reshape(box.numpy_shape)

        # 2. Select game window from entire screen.
        window_name = 'select_screen_area'
        region = cv2.selectROI(window_name, entire_screen)

        # `region` is tuple for (x1, y1, x2 - x1, y2 - y1) according to `Box`
        # coordinate system.
        return Box(
            region[0], region[1],
            region[0] + region[2], region[1] + region[3]
        )

    @staticmethod
    def get_local_monitors():
        # We currently use mss().monitors to get monitor information.
        # mss().monitors[0] is a dict of all monitors together
        # mss().monitors[N] is a dict of the monitor N (with N > 0)
        # But we drop the first elem because it's sometimes wrong because of
        # the bug in mss module.
        # TODO: Need to cache this property somewhere.
        mss_monitors = mss().monitors[1:]
        monitors = []
        for idx, elem in enumerate(mss_monitors):
            monitors.append(Monitor(
                elem['width'],
                elem['height'],
                elem['left'],
                elem['top'],
                idx == 0
            ))
        return monitors


class _LocalImpl(object):
    def __init__(self, box):
        self._box = box

    def read_screen(self):
        """Reads RGB screen data and returns corresponding 1 dimensional
        numpy array so that it can be reshaped later.

        """
        return self._post_process(
            self._read(self._box)
        )

    def _read(self, bounding_box):
        """Reads screen and returns raw RGB `bytearray`.

        :param bounding_box: Read only given area of the screen.

        """
        raise NotImplementedError()

    def _post_process(self, raw):
        """Parses `bytearray` to `numpy.ndarray`

        """
        single_byte_uint = '|u1'
        return np.frombuffer(raw, dtype=single_byte_uint)


class MssImpl(_LocalImpl):
    def __init__(self, box):
        super(MssImpl, self).__init__(box)
        self._is_osx = platform == 'darwin'
        self._executor = mss()

        if self._is_osx:
            # XXX: `mss` passes wrong param when it calls
            # `coregraphics.CGWindowListCreateImage' resulting in doubled
            # image size in retina display.
            # We fix this by hooking that function directly.
            orig_func = self._executor.core.CGWindowListCreateImage

            def _hook(screen_bounds, list_option, window_id, image_option):
                norminal_resolution = 1 << 4
                return orig_func(
                    screen_bounds, list_option, window_id,
                    norminal_resolution
                )
            self._executor.core.CGWindowListCreateImage = _hook

    def _read(self, bounding_box):
        """FIXME: Currently taking a screenshot is not fast enough in
        retina-like display which has a very high pixel density.

        """
        # Coordinates need to be converted accordingly.
        x1, y1, x2, y2 = bounding_box.to_tuple()
        width = x2 - x1
        height = y2 - y1

        monitor_dict = {
            'left': x1,
            'top': y1,
            'width': width,
            'height': height
        }

        adjust_needed = self._is_osx and width % 16 != 0
        if adjust_needed:
            # XXX: When the width is not divisible by 16, extra padding is
            # added by macOS in the form of black pixels, which results
            # in a screenshot with shifted pixels.
            # To prevent this, `mss` reduces width to the closest smaller
            # multiple of 16.
            # But we don't want the width size to be reduced unexpectedly.
            # This is a little hack to get the exact size of image.
            adjusted_width = width + (16 - (width % 16))
            monitor_dict['width'] = adjusted_width

            # Now `adjusted_rgb_data` has a bigger width.
            # Let's cut the remaining width to get our desired width.
            adjusted_rgb_data = self._executor.grab(monitor_dict).rgb
            rgb_data = bytearray()
            num_channels = 3
            for idx in range(height):
                offset = idx * (adjusted_width * num_channels)
                rgb_data.extend(
                    adjusted_rgb_data[offset:offset + width * num_channels]
                )

            return rgb_data
        else:
            return self._executor.grab(monitor_dict).rgb


class PilImpl(_LocalImpl):
    pass


class ScreenGrab(object):
    def __init__(self, box):
        """Constructor

        :param box: This class will capture screen inside this `Box`.

        """
        self._box = box

    @property
    def box(self):
        return self._box

    @box.setter
    def box(self, x):
        self._box = x

    def prepare(self):
        raise NotImplementedError()

    def close(self):
        raise NotImplementedError()

    @property
    def ready(self):
        raise NotImplementedError()

    def grab(self):
        raise NotImplementedError()

    def __enter__(self):
        return self

    def __exit__(self):
        self.close()

    def __repr__(self):
        return '<ScreenGrab [%s]>' % self._box.name


class LocalScreenGrab(ScreenGrab):
    def __init__(self, box, impl=MssImpl):
        self._impl = impl(box)

    def grab(self):
        return self._impl.read_screen()


def stream_local_game_screen(box=None, default_fps=10):
    """Convenient wrapper for local screen capture.
    This method wraps everything which is needed to get game screen data in
    primary monitor.
    It generates RGB numpy array with the shape of (height, width, 3)
    rather than raw 1 dimensional array because selected image size is not
    exposed to outside this method.

    :param box: If it's None, we first select game window area from screen
    and start streaming inside that box.
    :param default_fps: Target fps for screen capture. NOTE that this value
    can be adjusted from other coroutine.

    """
    if box is None:
        # Open a new process and get output from stdout
        # Due to bug in cv2.selectROI, area selection window sometime hangs
        # until main process exits preventing us from capturing right screen
        # data. We have to start subprocess to make sure the window to be
        # closed before we start capturing the screen.
        dir_path = os.path.dirname(os.path.realpath(__file__))
        subproc = Popen([
            'python', os.path.join(
                dir_path, '..', 'scripts/select_screen_area.py')
        ], stdout=PIPE)
        output, _ = subproc.communicate()

        try:
            box_tuple = ast.literal_eval(output.split('\n')[-2])
        except ValueError:
            # Something went wrong.
            traceback.print_exc()
            raise ScreenException('Failed to get screen area')

        box = Box.from_tuple(box_tuple)

    # We may need to use some epsilon value to meet fps more tightly.
    time_per_frame = 1.0 / default_fps
    local_grab = LocalScreenGrab(box)
    while True:
        start = time.time()

        screen = local_grab.grab()
        target_fps = yield screen.reshape(box.numpy_shape)
        if target_fps is not None:
            # Change fps accordingly
            time_per_frame = 1.0 / target_fps

        execution_time = time.time() - start
        if execution_time > time_per_frame:
            # Too high fps. No need to sleep.
            pass
        else:
            time.sleep(time_per_frame - execution_time)



"""
europilot.compat
~~~~~~~~~~~~~~~~

This module handles compatibility issues between Python 2 and 3.

"""

import sys

_ver = sys.version_info
is_py2 = _ver[0] == 2
is_py3 = _ver[0] == 3

if is_py2:
    from Queue import Queue
elif is_py3:
    from queue import Queue



from europilot.screen import Box
from europilot.train import generate_training_data, Config

class MyConfig(Config):
    # Screen area
    BOX = Box(0, 0, 500, 500)
    # Screen capture fps
    DEFAULT_FPS = 20

generate_training_data(config=MyConfig)



import threading

from europilot.controllerstate import ControllerState

if __name__ == '__main__':
    c = ControllerState()
    c.start()

    def print_output():
        threading.Timer(10.0, print_output).start()
        print(c.get_state())

    print_output()

    # ON_POSIX = 'posix' in sys.builtin_module_names
    # p = Popen(['python', '-u', 'g27.py'], bufsize=0, stdout=PIPE,
    #                close_fds=ON_POSIX)
    # for line in iter(p.stdout.readline, ''):
    #     print("line: %s" % line)
    # p.stdout.close()



from PIL import Image

from europilot.screen import ScreenUtils, LocalScreenGrab

# Select game screen
box = ScreenUtils.select_screen_area()

# Grab selected area
local_grab = LocalScreenGrab(box)
arr = local_grab.grab()
arr = arr.reshape(box.numpy_shape)

# Show image
image = Image.fromarray(arr, 'RGB')
image.show()



"""
europilot.train
~~~~~~~~~~~~~~~

Utils for generating training data.

"""

import os
import re
import time
import hashlib
import datetime
import threading
import multiprocessing

from PIL import Image
from pynput import keyboard

from europilot.compat import Queue
from europilot.exceptions import TrainException
from europilot.screen import stream_local_game_screen
from europilot.controllerstate import ControllerOutput


class _ConfigType(type):
    def __getattr__(self, attr):
        raise TrainException('Invalid configuration: %s' % attr)


class Config(object):
    """Default configuration.
    Anyone who wants to use custom configuration should inherit this class.

    :attr BOX: `screen.Box` object indicating capture area.
    :attr DATA_PATH: Csv file path.
    :attr IMG_PATH: Image file path.
    :attr IMG_EXT: Image file extension.
    :attr TRAIN_UID: If it's not None, file existing csv file with a name
    starts with `train_uid` and add data to that file.
    :attr DEFAULT_FPS: Default target fps for screen capture.
    :attr WAIT_KEYPRESS: If this is True, do not start training until start
    key (keyboard `r`, wheel `left button 1`) is pressed.
    :attr DEBUG: If this is True, write debug msg to stdout.

    """
    __metaclass__ = _ConfigType
    BOX = None
    DATA_PATH = os.path.join('data', 'csv')
    IMG_PATH = os.path.join('data', 'img', 'raw')
    IMG_EXT = 'jpg'
    TRAIN_UID = None
    DEFAULT_FPS = 10
    WAIT_KEYPRESS = False
    DEBUG = True


_WORKER_BREAK_FLAG = 'stop_consuming'


def _print(text):
    if _global_config.DEBUG:
        print(text)


class Worker(multiprocessing.Process):

    def __init__(self, train_uid, inq, outq):
        """
        :param train_uid: This would be image filename prefix.
        :param inq: inbound multiprocessing.Queue (main process <-> worker)
        :param outq: outbound multiprocessing.Queue (worker <-> writer)

        """
        multiprocessing.Process.__init__(self)
        self._train_uid = train_uid
        self._inq = inq
        self._outq = outq
        self._img_path = _global_config.IMG_PATH
        self._img_ext = _global_config.IMG_EXT
        self.daemon = True

    @property
    def train_uid(self):
        return self._train_uid

    def run(self):
        while True:
            try:
                # Blocking `get`
                data = self._inq.get()

                if data == _WORKER_BREAK_FLAG:
                    break

                image_data, sensor_data = data
                filename = self._save_image(image_data)
                self._outq.put((filename, sensor_data))
            except KeyboardInterrupt:
                pass

    def _save_image(self, image_data):
        """Synchronously write image data to disk and returns filename.
        :param image_data: RGB numpy array

        """
        # filename example: '{self._train_uid}_2017_07_24_21_18_46_13.jpg'
        filename = self._train_uid + '_' + re.sub(
            '[-:.]', '_', datetime.datetime.now().isoformat('_')[:-4]) + \
            '.' + self._img_ext

        image = Image.fromarray(image_data, 'RGB')
        image.save(os.path.join(self._img_path, filename))
        return filename


class Writer(multiprocessing.Process):
    def __init__(self, train_uid, inq, csv_initialized=False):
        multiprocessing.Process.__init__(self)
        self._train_uid = train_uid
        self._inq = inq
        self._data_path = _global_config.DATA_PATH
        self._data_seq = 0
        self._csv_initialized = csv_initialized
        self._filename = self._train_uid + '.csv'

    @property
    def filename(self):
        return self._filename

    def run(self):
        f = os.path.join(self._data_path, self._filename)
        with open(f, 'a' if os.path.isfile(f) else 'w') as file_:
            while True:
                try:
                    data = self._inq.get()
                    if data == _WORKER_BREAK_FLAG:
                        break
                    image_filename, sensor_data = data

                    self._write(file_, image_filename, sensor_data)

                    if self._data_seq % 10 == 0:
                        _print('seq: %s, filename: %s, datetime: %s' %
                            (
                                self._data_seq,
                                image_filename,
                                str(datetime.datetime.now())
                            )
                        )
                except KeyboardInterrupt:
                    pass

    def _write(self, file_, image_filename, sensor_data):
        """Synchronously write sensor data to disk.
        :param image_filename: filename of corresponding training image
        :param sensor_data: `SensorData` object.

        CSV format
        seq_id,filename,sensor_data1,sensor_data2,sensor_data3,...

        Order of sensor data depends on `OrderedDict` defined in
        `controllerstate.ControllerState`.

        """
        if not self._csv_initialized:
            # Add headers
            sensor_header = ','.join(sensor_data.raw.keys())
            csv_header = 'img,' + sensor_header
            file_.write(csv_header + '\n')
            self._csv_initialized = True

        values = [image_filename] + [str(x) for x in sensor_data.raw.values()]
        data = ','.join(values)
        self._data_seq += 1
        file_.write(data + '\n')


_train_sema = threading.BoundedSemaphore(value=1)
_global_config = Config


def generate_training_data(config=Config):
    """Generate training data.

    :param config: Training configuration class

    """
    try:
        # Set global config so that other workers can access it directly.
        global _global_config
        _global_config = config

        # Check if data paths exist and are writable.
        if not os.access(config.DATA_PATH, os.W_OK):
            raise TrainException('Invalid data path: %s' % config.DATA_PATH)
        if not os.access(config.IMG_PATH, os.W_OK):
            raise TrainException('Invalid image path: %s' % config.IMG_PATH)

        streamer = stream_local_game_screen(
            box=config.BOX, default_fps=config.DEFAULT_FPS)

        worker_q = multiprocessing.Queue()
        writer_q = multiprocessing.Queue()
        num_workers = multiprocessing.cpu_count()
        workers = []

        train_uid = config.TRAIN_UID
        csv_initialized = train_uid is not None
        if train_uid is None:
            # Generate train_uid to start new data generation.
            # Call `encode` because hashlib in python3 doesn't accept unicode.
            d = str(datetime.datetime.now()).encode('utf8')
            train_uid = hashlib.md5(d).hexdigest()[:8]

        for i in range(num_workers):
            workers.append(
                Worker(train_uid, worker_q, writer_q)
            )

        # NOTE that these workers are daemonic processes
        for worker in workers:
            worker.start()

        # Start 1 process
        writer = Writer(train_uid, writer_q, csv_initialized)
        writer.start()

        # Start 1 thread
        control_q = Queue()
        flow_controller = FlowController(control_q)
        flow_controller.start()

        # Start 1 process and 1 thread.
        def _state_listener(sensor_data):
            _feed_control_signal(control_q, sensor_data=sensor_data)
        controller_output = ControllerOutput(state_listener=_state_listener)
        controller_output.start()

        # Start 1 thread
        keyboard_listener = KeyListener(control_q)
        keyboard_listener.start()

        fps_adjuster = FpsAdjuster()
        last_sensor_data = None

        if config.WAIT_KEYPRESS:
            if config.BOX is None:
                # Select area and drop first screen image.
                next(streamer)
            _train_sema.acquire()

        while True:
            if flow_controller.acquired:
                # Switch context and give flow_controller time to acquire sema.
                time.sleep(0.1)

            # TODO: Do something to resolve this in python 2.x
            # https://bugs.python.org/issue8844
            _train_sema.acquire()

            if last_sensor_data is None:
                # Start generator
                image_data = next(streamer)
            else:
                image_data = streamer.send(
                    fps_adjuster.get_next_fps(last_sensor_data))

            sensor_data = controller_output.get_latest_state_obj()
            last_sensor_data = sensor_data
            worker_q.put((image_data, sensor_data))

            try:
                _train_sema.release()
            except ValueError:
                pass
    except KeyboardInterrupt:
        try:
            _train_sema.release()
        except ValueError:
            # Already released.
            pass

        control_q.put(_WORKER_BREAK_FLAG)
        controller_output.terminate()

        # Gracefully stop workers
        for _ in range(num_workers):
            worker_q.put(_WORKER_BREAK_FLAG)
        writer_q.put(_WORKER_BREAK_FLAG)

        for worker in workers:
            worker.terminate()

        for worker in workers:
            worker.join()

        writer.terminate()
        writer.join()


def _feed_control_signal(q, key_value=None, sensor_data=None):
    """Parse device (keyboard, wheel) input to control signal and
    feed it into `FlowController` thread.
    If both devices put input simultaneously, key_value is ignored.

    :param q: `Queue` between main thread and `FlowController` thread.
    :param key_value: Keyboard input
    :param sensor_data: `SensorData` object

    """
    control_signal = None

    if key_value is not None:
        if key_value == 'r':
            control_signal = FlowController.RESUME_SIGNAL
        elif key_value == 'q':
            control_signal = FlowController.PAUSE_SIGNAL

    if sensor_data is not None:
        if sensor_data.resume_button_pressed:
            control_signal = FlowController.RESUME_SIGNAL
        elif sensor_data.pause_button_pressed:
            control_signal = FlowController.PAUSE_SIGNAL

    if control_signal is not None:
        q.put(control_signal)


class KeyListener(keyboard.Listener):
    def __init__(self, outq):
        """Constructor.
        :param outq: `Queue` between this thread and `FlowController` thread.

        `keyboard.Listener` is a daemon thread by default.

        """
        self._outq = outq
        super(KeyListener, self).__init__(on_press=self._on_press)

    def _on_press(self, key):
        try:
            _feed_control_signal(self._outq, key_value=key.char)
        except AttributeError:
            # Pressed key is not alphabet.
            pass


class FlowController(threading.Thread):
    """This thread controls data generation process by parsing signal from
    input devices.

    """
    PAUSE_SIGNAL = 'pause'
    RESUME_SIGNAL = 'resume'

    def __init__(self, inq):
        threading.Thread.__init__(self)
        self.daemon = True
        self._inq = inq
        self._acquired = False

    @property
    def acquired(self):
        return self._acquired

    def run(self):
        while True:
            signal = self._inq.get()
            if signal == self.RESUME_SIGNAL:
                self._resume_data_generation()
            elif signal == self.PAUSE_SIGNAL:
                self._pause_data_generation()
            elif signal == _WORKER_BREAK_FLAG:
                break

    def _pause_data_generation(self):
        _print('Data generation paused')
        if not self._acquired:
            self._acquired = True
            _train_sema.acquire()

    def _resume_data_generation(self):
        _print('Data generation resumed')
        try:
            _train_sema.release()
            self._acquired = False
        except ValueError:
            pass


class FpsAdjuster(object):
    def __init__(self):
        self._default_fps = _global_config.DEFAULT_FPS
        self._adjust_factor = 2
        self._duration_threshold = 2
        self._max_straight_wheel_axis = 10
        self._last_straight_time = None

    def get_next_fps(self, sensor_data):
        """Adjust fps according to wheel axis
        We want to reduce fps when the car is going straight longer than
        threshold seconds.

        :param sensor_data: `controllerstate.SensorData` object.

        """
        going_straight = self._going_straight(sensor_data.wheel_axis)
        if self._last_straight_time is None:
            self._update_last_straight_time(going_straight)
            return self._default_fps

        straight_duration = time.time() - self._last_straight_time
        if going_straight and \
                straight_duration > self._duration_threshold:
            # Adjust fps
            return max(self._default_fps / self._adjust_factor, 1)
        else:
            self._update_last_straight_time(going_straight)
            return self._default_fps

    def _going_straight(self, wheel_axis):
        return abs(int(wheel_axis)) < self._max_straight_wheel_axis

    def _update_last_straight_time(self, going_straight):
        if self._last_straight_time is None and going_straight:
            self._last_straight_time = time.time()
        elif not going_straight:
            self._last_straight_time = None
        # If last_controller_state is not None and going_straight, do nothing



"""Select screen area with opencv and write output to stdout.
This script will be executed from `europilot.screen.stream_local_game_screen`.

"""
import os
import sys
dir_path = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(dir_path, '..'))

from europilot.screen import ScreenUtils

# Select screen area and write selected output to stdout
box = ScreenUtils.select_screen_area()
print(box.to_tuple())



import os
import sys
import json
from subprocess import PIPE, Popen
from threading import Thread
from collections import OrderedDict


class ControllerOutput(object):
    """Reads stdout from joystick.py, to get the controller event outputs without
    blocking.

    """
    def __init__(self, state_listener=None):
        """Constructor.

        :param state_listener: Listener which takes single `SensorData` param.

        """
        self._controller_state = ControllerState()
        self._state_listener = state_listener

    def start(self):
        ON_POSIX = 'posix' in sys.builtin_module_names
        # use unbuffered output, since stdout will be generally small
        # https://stackoverflow.com/questions/1410849/bypassing-buffering-of-subprocess-output-with-popen-in-c-or-python
        dir_path = os.path.dirname(os.path.realpath(__file__))

        # XXX: This `python` may not be our desired python bin
        self.p = Popen(
            ['python', '-u', os.path.join(dir_path, 'joystick.py')],
            bufsize=0,
            stdout=PIPE,
            close_fds=ON_POSIX
        )
        self.t = Thread(target=self.__update_state,
                        args=(self.p,))
        self.t.daemon = True  # thread dies with program
        self.t.start()

    def terminate(self):
        self.p.terminate()

    def get_latest_state_obj(self):
        return self._controller_state.get_state_obj()

    def __update_state(self, process):
        for line in iter(process.stdout.readline, ''):
            self._controller_state.update_state(line.strip())
            self._state_listener(self._controller_state.get_state_obj())

        process.stdout.close()


class ControllerState(object):
    """Thread-safe controller state.
    Holds the latest value of each of the controller output value.
        >>> c = ControllerState()
        >>> c.start()
        >>> c.get_state()
        OrderedDict([('wheel-axis', '1012'), ('clutch', '-27865'),...]
    """
    def __init__(self):
        self.state = self.__init_dict()

    def __init_dict(self):
        """Initialize the values for each of the controller output"""
        d = OrderedDict()
        d['wheel-axis'] = '0'
        d['clutch'] = '0'
        d['brake'] = '0'
        d['gas'] = '0'
        d['paddle-left'] = '0'
        d['paddle-right'] = '0'
        d['wheel-button-left-1'] = '0'
        d['wheel-button-left-2'] = '0'
        d['wheel-button-left-3'] = '0'
        d['wheel-button-right-1'] = '0'
        d['wheel-button-right-2'] = '0'
        d['wheel-button-right-3'] = '0'
        d['shifter-button-left'] = '0'
        d['shifter-button-right'] = '0'
        d['shifter-button-up'] = '0'
        d['shifter-button-down'] = '0'
        d['dpad-left/right'] = '0'
        d['dpad-up/down'] = '0'
        d['shifter-button-1'] = '0'
        d['shifter-button-2'] = '0'
        d['shifter-button-3'] = '0'
        d['shifter-button-4'] = '0'
        d['gear-1'] = '0'
        d['gear-2'] = '0'
        d['gear-3'] = '0'
        d['gear-4'] = '0'
        d['gear-5'] = '0'
        d['gear-6'] = '0'
        d['gear-R'] = '0'

        return d

    def update_state(self, msg):
        """Update ControllerState with the latest controller data"""
        k, v = msg.split()
        if k in self.state:
            self.state[k] = v

    def get_state(self):
        """Returns the latest state"""
        return self.state

    def get_state_obj(self):
        """Returns the latest `SensorData` object"""
        return SensorData.from_ordered_dict(self.get_state())

    def get_state_json(self):
        """Returns the latest state in json format"""
        j = json.dumps(self.state)
        return j


class SensorData(object):
    def __init__(self, data):
        """Constructor.
        :param data: `OrderedDict` containing sensor data.

        """
        self._data = data

    @staticmethod
    def from_ordered_dict(dict_):
        return SensorData(dict_)

    @property
    def raw(self):
        return self._data

    @property
    def wheel_axis(self):
        return self._data['wheel-axis']

    @property
    def resume_button_pressed(self):
        return self._data['wheel-button-right-1'] == '1'

    @property
    def pause_button_pressed(self):
        return self._data['wheel-button-left-1'] == '1'



import os
from setuptools import setup


install_requires = [
    'mss==3.0.1',
    'numpy>=1.12.1',
    'Pillow>=4.2.1',
    'pynput>=1.3.5'
]

about = {}
cur_path = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(
        cur_path, 'europilot', '__version__.py'), 'r') as f:
    exec(f.read(), {'__builtins__' : None}, about)

setup(
    name=about['__title__'],
    version=about['__version__'],
    description=about['__description__'],
    packages=['europilot', 'scripts'],
    install_requires=install_requires
)



"""
europilot.exceptions
~~~~~~~~~~~~~~~~~~~~

Defines custom exceptions

"""


class EuropilotException(Exception):
    pass


class ScreenException(EuropilotException):
    pass


class TrainException(EuropilotException):
    pass



"""
http://upgrayd.blogspot.de/2011/03/logitech-dual-action-usb-gamepad.html

G27
===
example::

    A0 B7 A3 04 5C 7D 02 02
     0  1  2  3  4  5  6  7


    0, 1, 2, 3: sequence, little endian
    4, 5: value, little endian
    6: group
    7: axis

NOTE! From here on, I talk big endian -- also in hex!

Wheel values::

            left               dead           right
            <-------------------XX---------------->
    dec     32769      65535     0      1     32767
    hex     80 01      ff ff  00 00 00 01     7F FF


Pedal values::

- no pressure: 7F FF
- halfway: 00 00
- full: 80 01


Button values::

- on: 01 00
- off: 00 00


Gear values::

- on: 01 00
- off: 00 00


Arrow Pad values::

- left/up: 01 80 (32769)
- right/down: ff 7f (32767)
- off : 00 00


For wheel values, the output range is normalized to [-32767, 32767].
For pedal values, the output range is normalized to [0, 65535].
For Arrow Pad values, the output range is normalized to [-1, 1].


"""
import os
import sys
import time
import random
from binascii import hexlify

BUTTON_NAME = """
0200=wheel-axis
0201=clutch
0203=brake
0202=gas
0105=paddle-left
0104=paddle-right
0107=wheel-button-left-1
0114=wheel-button-left-2
0115=wheel-button-left-3
0106=wheel-button-right-1
0112=wheel-button-right-2
0113=wheel-button-right-3
0101=shifter-button-left
0102=shifter-button-right
0103=shifter-button-up
0100=shifter-button-down
0204=dpad-left/right
0205=dpad-up/down
010b=shifter-button-1
0108=shifter-button-2
0109=shifter-button-3
010a=shifter-button-4
010c=gear-1
010d=gear-2
010e=gear-3
010f=gear-4
0110=gear-5
0111=gear-6
0116=gear-R
"""


class Bytewurst(object):

    def __init__(self, bs):
        self.raw = bs
        self.ints = map(ord, bs)

    def __repr__(self):
        return ' '.join(map(hexlify, self.raw))

    @property
    def int(self):
        """
        For "01 0A" ints would be [1, 10], so::
            >>> bs = '\x01\x0A'
            >>> bw = Bytewurst(bs)
            >>> bw.int == (1 * 1) + (10 * 256)
        """

        def powergenerator(start=0):
            """Generate powers of 256"""
            i = start
            while True:
                yield 256 ** i
                i += 1

        return sum(a * b for a, b in zip(self.ints, powergenerator()))

    @property
    def hexLE(self):
        return hexlify(self.raw)

    @property
    def bits(self):
        return ' '.join([format(x, '08b') for x in self.ints])


class Button(Bytewurst):
    def __init__(self, bs):
        super(Button, self).__init__(bs)
        button_namedict = dict(line.split('=') for line in
                               BUTTON_NAME.strip().split('\n'))
        self.name = button_namedict.get(self.hexLE, 'UNKNOWN:%s' % self.hexLE)


class Value(Bytewurst):
    def __repr__(self):
        if self.int == 0:
            return ' off'
        elif self.int == 1:
            return ' on'
        else:
            return ' %d' % self.int

    def int_normalized(self, name):
        """ Normalizes value to an adequate range

        For wheel values, the output range is normalized to [-32767, 32767].
        For pedal values, the output range is normalized to [0, 65535].
        For Arrow Pad values, the output range is normalized to [-1, 1].
        """
        v = super(Value, self).int
        if name == 'wheel-axis':
            if v >= 32769:
                v = v - 65536
        elif name == 'clutch' or name == 'brake' or name == 'gas':
            if v >= 32769:
                v = -v + 98304
            else:
                v = -v + 32767
        elif name == 'dpad-left/right' or name == 'dpad-up/down':
            if v == 32769:
                v = -1
            elif v == 32767:
                v = 1

        return v


class Message(object):
    def __init__(self, bs):
        self.bs = bs
        self.raw_seq = bs[0:4]
        self.raw_value = bs[4:6]
        self.raw_id = bs[6:8]
        self.ints = map(ord, bs)
        self.sequence = Bytewurst(bs[0:4])
        self.value = Value(bs[4:6])
        self.button = Button(bs[6:8])

    def __repr__(self):
        values = (self.button.name,
                  self.value.int_normalized(self.button.name))
        return ' '.join(map(str, values))


class LinuxVirtualJoystick(object):
    """
    A Virtual joystick driver is implemented by attaching userspace device
    drivers in the kernel by outputting events into udev.

    Relevant information about the virtual driver was derived with evtest.

    A LinuxVirtualJoystick object must be initialized before ETS is started
    for the game to recognize the virtual driver.

    Note that you may have to adjust the control settings inside ETS to
    map the events with the necessary values. Check the project page for more
    details.
    """
    def __init__(self, name='Virtual G27 Racing Wheel', bustype=0x0003,
                 vendor=0x046d, product=0xc29b, version=0x0111, events=None):
        if events is None:
            events = (
                # EV_KEY
                uinput.BTN_TRIGGER,
                uinput.BTN_THUMB,
                uinput.BTN_THUMB2,
                uinput.BTN_TOP,
                uinput.BTN_TOP2,
                uinput.BTN_PINKIE,
                uinput.BTN_BASE,
                uinput.BTN_BASE2,
                uinput.BTN_BASE3,
                uinput.BTN_BASE4,
                uinput.BTN_BASE5,
                uinput.BTN_BASE6,
                (0x01, 0x12c),  # (event code 300)
                (0x01, 0x12d),  # (event code 301)
                (0x01, 0x12e),  # (event code 302)
                uinput.BTN_DEAD,
                uinput.BTN_TRIGGER_HAPPY1,
                uinput.BTN_TRIGGER_HAPPY2,
                uinput.BTN_TRIGGER_HAPPY3,
                uinput.BTN_TRIGGER_HAPPY4,
                uinput.BTN_TRIGGER_HAPPY5,
                uinput.BTN_TRIGGER_HAPPY6,
                uinput.BTN_TRIGGER_HAPPY7,

                # EV_ABS
                uinput.ABS_X + (-32767, 32767, 0, 0),   # steering wheel
                uinput.ABS_Y + (0, 65535, 0, 0),   # clutch
                uinput.ABS_Z + (0, 65535, 0, 0),   # acceleration
                uinput.ABS_RZ + (0, 65535, 0, 0),  # break
                uinput.ABS_HAT0X + (-1, 1, 0, 0),
                uinput.ABS_HAT0Y + (-1, 1, 0, 0),
            )

        self.device = uinput.Device(events,
                                    name=name,
                                    bustype=bustype,
                                    vendor=vendor,
                                    product=product,
                                    version=version)

    def emit(self, angle, clutch=None, accel=None, brk=None):
        # emit as one atomic event, by using syn=True at the last emit call
        # for more information check the source code at
        # https://github.com/tuomasjjrasanen/python-uinput/blob/master/src/__init__.py#L191

        if clutch:
            self.device.emit(uinput.ABS_Y, clutch, syn=False)
        if accel:
            self.device.emit(uinput.ABS_Z, accel, syn=False)
        if brk:
            self.device.emit(uinput.ABS_RZ, brk, syn=False)

        self.device.emit(uinput.ABS_X, angle, syn=True)

    def __del__(self):
        self.device.destroy()


# Handle platform dependent joystick impl
if sys.platform.startswith('linux'):
    import uinput
    VirtualJoystick = LinuxVirtualJoystick
elif sys.platform == 'darwin':
    pass


if __name__ == '__main__':
    def _dump_messages(input_):
        with open(input_, 'rb') as device:
            while True:
                bs = device.read(8)
                message = Message(bs)
                print(message)

    def _dump_dummy_messages():
        # Need to simulate straight wheel axis to test realtime fps adjustment.
        straight = False
        straight_start_time = None
        straight_duration = 5
        while True:
            if not straight and random.randint(0, 99) == 0:
                # Straight wheel axis for 5 secs.
                straight = True
                straight_start_time = time.time()

            if straight and \
                    time.time() - straight_start_time > straight_duration:
                # End of straight wheel axis
                straight = False
                straight_start_time = None

            # Axis value 0 will be regarded as straight wheel regardless of
            # `train.FpsAdjuster._max_straight_wheel_axis`.
            wheel_axis = random.randint(-32767, 32767) if not straight else 0

            print('wheel-axis ' + str(wheel_axis))

    # TODO: Get device path as argument
    device = '/dev/input/js0'
    if os.path.exists(device):
        _dump_messages(device)
    else:
        # When joystick doesn't exist. Let's dump dummy messages.
        # TODO: Warn this to stdout so that we can be aware of mock g27.
        _dump_dummy_messages()



import time

from europilot.screen import stream_local_game_screen, Box

box = Box(0, 0, 500, 500)
streamer = stream_local_game_screen(box=box)
while True:
    arr = next(streamer)
    print((type(arr), arr.shape))
    # Do something with arr
    time.sleep(0.5)

